
# compare objects by value not identity
class NaturalEqualityMixin(object):
    def __eq__(self, other):
        if type(self) is type(other):
            return self.__dict__ == other.__dict__
        return False

    def __ne__(self, other):
        return not self.__eq__(other)

def equals(lhs, rhs):
    assert(is_node(lhs) and is_node(rhs))
    return lhs.__dict__ == rhs.__dict__

# generic node representation
class TreeNode(object):
    def __init__(self, **kwargs):
        # group can be: 'ast-node', 'annotation', 'encoding', atoms (non dict's)
        self.group = 'ast-node'
        self.__dict__.update(kwargs)
        if not hasattr(self, 'child_keys'):
            self._rebuilt_children()

    # If value is TreeNode or a sequence with atleast a TreeNode (normally
    # should be only TreeNodes) then add it's name (key) as child_key
    # (allows simple automating complete tree visiting).
    def _rebuilt_children(self):
        self.child_keys = []
        for k,v in self.__dict__.iteritems():
            if isinstance(v, TreeNode):
                self.child_keys.append(k)
            elif isinstance(v, (list, set, frozenset, tuple)):
                if any(isinstance(e, TreeNode) for e in v):
                    self.child_keys.append(k)
        return self

# different node constructors
def _create_node(meta, *args, **kwargs):
    if len(args) > 2:
        raise RuntimeError('to many positional arguments supplied')
    elif len(args) < 1:
        raise RuntimeError('at least the tag should be supplied')
    tag = args[0]
    sloc = None
    try:
        sloc = args[1]
    except IndexError:
        pass
    kwargs.update(meta)
    kwargs.update({'tag': tag, 'sloc': sloc})
    return TreeNode(**kwargs)

def Expr(*args, **kwargs):
    return _create_node({'cls': 'expr',
                         'lvalue': False}, *args, **kwargs)

def Stmt(*args, **kwargs):
    return _create_node({'cls': 'stmt'}, *args, **kwargs)

def Decl(*args, **kwargs):
    return _create_node({'cls': 'decl'}, *args, **kwargs)

def Type(*args, **kwargs):
    return _create_node({'cls': 'type'}, *args, **kwargs)

def Misc(*args, **kwargs):
    return _create_node({'cls': 'misc'}, *args, **kwargs)


def is_node(node): return isinstance(node, TreeNode)
def is_expr(node): return is_node(node) and node.cls == 'expr'
def is_stmt(node): return is_node(node) and node.cls == 'stmt'
def is_decl(node): return is_node(node) and node.cls == 'decl'
def is_type(node): return is_node(node) and node.cls == 'type'
