# generic (right now unordered due to unordered child-keys) tree visitor
# (for ivl the order is unimportent but it might be necessary for other
# languages)
def _identity(x):
    return x

def _const(x):
    return lambda a: a

def _fnone(fun, node):
    x = fun(node)
    return node if x is None else x

class AbortWalk(object):
    def __init__(self, result=None):
        self.result = result

def _should_abort(result):
    return isinstance(result, AbortWalk)

class Args(object):
    def __init__(self, _result=None, **kwargs):
        self.result = _result
        self.kwargs = kwargs

def _has_kwargs(result):
    return isinstance(result, Args)

# down can return:
#  - None
#  - (node, bool)
#  - node
#  - bool -> abort deeper visiting
# up can return:
#  - None
#  - node
def walk(node, down, up):
    if hasattr(node, 'child_keys'):
        # node seems to be a TreeNode -> so visit it
        d = _fnone(down, node)
        if _should_abort(d):
            # don't walk down any further, take result if not None else keep node
            d = _fnone(_const(d.result), node)
        else:
            # continue walking -> visit all children
            for e in d.child_keys:
                setattr(d, e, walk(getattr(d, e), down, up))
        return _fnone(up, node)
    else:
        # assume node is a sequence
        return type(node)(walk(n, down, up) for n in node)

def walk_down(node, down):
    return walk(node, down, _identity)

def walk_up(node, up):
    return walk(node, _identity, up)



# tree walker base classes
# The default implementation allows simple dispatching based on classname
# argument. More sophisticated approuch could use some form of predicate
# dispatch.
    
class WalkUp(object):
    def leave(self, node):
        """Called for each node while walking Up."""
        pass
##        return getattr(self, '_leave_' + node.classname)(node)

    def run(self, root):
        return walk_up(root, self.leave)

class WalkDown(object):
    def enter(self, node):
        """Called for each node while walking down."""
        pass
##        return getattr(self, '_enter_' + node.classname)(node)

    def run(self, root):
        return walk_down(root, self.enter)

class Walker(WalkDown, WalkUp):
    def run(self, root):
        return walk(root, self.enter, self.leave)

