
class Error(Exception):
    pass

class DeclarationResolveError(Error):
    def __init__(self, name, sloc):
        self.name = name
        self.sloc = sloc
        super(DeclarationResolveError, self).__init__("unable to resolve '{}'".format(name))

class DoubleDeclarationError(Error):
    def __init__(self, name, sloc, prev_sloc):
        self.name = name
        self.sloc = sloc
        self.prev_sloc = prev_sloc
        super(DoubleDeclarationError, self).__init__("double declaration of '{}'".format(name))

# Variables, Events and Functions should exist in different namespaces.
# -> Need seperate Scopes for:
#    - Variables
#    - Events
#    - Functions / Threads / Updates
class Env(object):
    def __init__(self, parent=None, **kwargs):
        self.parent = parent
        self.decls = {}

    def resolve(self, name, sloc=None):
        try:
            return self.decls[name]
        except KeyError as e:
            if self.parent is None:
                raise DeclarationResolveError(name, sloc)
            return self.parent.resolve(name)

    def add(self, name, decl):
        if name in self.decls:
            raise DoubleDeclarationError(name, decl.sloc, self.decls[name].sloc)
        self.decls[name] = decl


def add_decl(env, decl):
    env.add(decl.name, decl)

def extend(env, decl_sequence):
    for e in decl_sequence:
        add_decl(env, e)

def resolve(env, name, sloc, expected=None):
    decl = env.resolve(name, sloc)
    if expected:
        if decl.tag != expected:
            DeclarationResolveError("unable to resolve '{}' named '{}'".format(expected, name), sloc)
    return decl